# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2020-2024 Luke Horwell <code@horwell.me>
"""
This module is a higher level interface for custom effects for devices that
support it, regardless of vendor.

This is used internally for rendering software effects as well as "scripted"
effects written by users.

To achieve effects, colours may be converted between different colour spaces:
    - RGB (Red, Green, Blue)            -> using hex codes for storage
    - HSV (Hue, Saturation, Value)      -> for colour intensity
    - HSL (Hue, Saturation, Lightness)  -> for black/white intensity
"""

import colour
import math

from . import common


class FX(object):
    """
    Backends use this class for the get_device_object() functionality. This
    is for supporting individual LED lighting where supported by the device.

    This object is the 'fx' object used in custom effect scripts as well as
    Polychromatic internally for software effects.

    All classes stubbed as NotImplementedError should be implemented.
    """
    def __init__(self):
        self.name = "Unknown Device"
        self.form_factor_id = "unrecognised"
        self.rows = 0
        self.cols = 0

    #######################################################
    # To be implemented by the backend
    #   See also: _backend.DeviceItem.Matrix()
    #######################################################
    def init(self):
        """
        Prepare the device for custom frames. If unnecessary, this can be ignored.
        """
        return

    def set(self, x=0, y=0, red=255, green=255, blue=255):
        """
        Set a colour at the specified co-ordinate.
        """
        raise NotImplementedError

    def draw(self):
        """
        Send the data to the hardware to be displayed.
        """
        raise NotImplementedError

    def clear(self):
        """
        Reset all LEDs to an off state.
        """
        raise NotImplementedError

    def brightness(self, percent):
        """
        Set the global brightness of the LEDs. Could be used for fade effects globally.
        """
        raise NotImplementedError

    #######################################################
    # Functions for scripting
    #######################################################
    def rgb_to_hex(self, red, green, blue):
        """
        Converts the specified RGB values into a HEX string.

        Input:  (0, 255, 0)
        Output: "#00FF00"
        """
        return common.rgb_to_hex([red, green, blue])

    def hex_to_rgb(self, value):
        """
        Converts a HEX colour string to a RGB list.

        Input:  "#FF0000"
        Output: [255, 0, 0]
        """
        return common.hex_to_rgb(value)

    def saturate_hex(self, hex_value, amount):
        """
        Change the saturation of a colour by a relative amount (-1 to 1) and
        return the new colour hex value.

        Params:
            hex_value   (str)   Hex value
            amount      (float) Relative amount to saturate (-1 to 1)
        """
        c = colour.Color(hex_value)

        # Value must be between 0 and 1
        new_value = c.get_saturation() + amount
        if new_value < 0:
            new_value = 0
        elif new_value > 1:
            new_value = 1

        c.set_saturation(new_value)
        return c.get_hex_l()

    def saturate_rgb(self, rgb, amount):
        """
        Alias to saturate_hex(), but inputs/outputs RGB values.

        Params:
            rgb         (list)  List of RGB integers [red, green, blue]
            amount      (float) Relative amount to saturate (-1 to 1)
        """
        return self.hex_to_rgb(self.saturate_hex(self.rgb_to_hex(rgb[0], rgb[1], rgb[2]), amount))

    def hue_hex(self, hex_value, amount):
        """
        Change the hue of a colour by a relative amount (-1 to 1) and
        return the new colour hex value.

        Params:
            hex_value   (str)   Hex value
            amount      (float) Relative amount to cycle hue (-1 to 1)
        """
        c = colour.Color(hex_value)
        c.set_hue(c.get_hue() + amount)
        return c.get_hex_l()

    def hue_rgb(self, rgb, amount):
        """
        Alias to hue_hex(), but inputs/outputs RGB values.

        Params:
            rgb         (list)  List of RGB integers [red, green, blue]
            amount      (float) Relative amount to cycle hue (-1 to 1)
        """
        return self.hex_to_rgb(self.hue_hex(self.rgb_to_hex(rgb[0], rgb[1], rgb[2]), amount))

    def lightness_hex(self, hex_value, amount):
        """
        Change the lightness of a colour by a relative amount (-1 to 1) and
        return the new colour hex value.

        Params:
            hex_value   (str)   Hex value
            amount      (float) Relative amount to cycle hue (-1 to 1)
        """
        c = colour.Color(hex_value)

        # Value must be between 0 and 1
        new_value = c.get_luminance() + amount
        if new_value < 0:
            new_value = 0
        elif new_value > 1:
            new_value = 1

        c.set_luminance(new_value)
        return c.get_hex_l()

    def lightness_rgb(self, rgb, amount):
        """
        Alias to lightness_hex(), but inputs/outputs RGB values.

        Params:
            rgb         (list)  List of RGB integers [red, green, blue]
            amount      (float) Relative amount to cycle hue (-1 to 1)
        """
        return self.hex_to_rgb(self.lightness_hex(self.rgb_to_hex(rgb[0], rgb[1], rgb[2]), amount))

    def gradient(self, colours=[], steps=0):
        """
        Return a list of colours that builds a gradient from start to finish
        using the specified colours (equal distance between each step)

        For best results, make sure the steps are divisable by the number
        of colours. Otherwise, you may will get more colours then expected.

        For example, ["#000000", "#FFFFFF"] (black, white) across 10 steps
        makes the 5th colour the midpoint (grey)

        Params:
            colours     (list)  List of colours for the gradient
            steps       (int)   Total colours to return for rendering the gradient
        """
        if len(colours) < 2:
            raise ValueError("Insufficient colours! At least 2 required to generate gradient.")

        output = []
        steps_between_stops = math.ceil(steps / (len(colours) - 1))

        for index, item in enumerate(colours):
            try:
                next_colour = colours[index + 1]
            except IndexError:
                # This is the last colour
                continue

            c = colour.Color(item)
            for c2 in list(c.range_to(next_colour, steps_between_stops)):
                output.append(c2.get_hex_l())

        return output



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2017-2024 Luke Horwell <code@horwell.me>
"""
Deprecated module for storing functions commonly used across Polychromatic's
interfaces and some backends.
"""
import colorama
import hashlib
import os
import sys
import subprocess
import grp
import time
import traceback
from threading import Thread

from . import base
from . import locales

# TODO: Refactor later!
paths = None

FORM_FACTORS = [
    "accessory",
    "display",
    "fan",
    "gpu",
    "headset",
    "keyboard",
    "keypad",
    "laptop",
    "mouse",
    "mousemat",
    "ram",
    "speaker",
    "stand",
    "unrecognised"
]



class Debugging(object):
    """
    Outputs pretty debugging details to the terminal.
    """
    def __init__(self):
        self.verbose_level = 0
        colorama.init()

        # Colours for stdout
        self.error = colorama.Fore.LIGHTRED_EX
        self.success = colorama.Fore.LIGHTGREEN_EX
        self.warning = colorama.Fore.YELLOW
        self.action = colorama.Fore.LIGHTYELLOW_EX
        self.blue = colorama.Fore.BLUE
        self.magenta = colorama.Fore.MAGENTA
        self.debug = colorama.Fore.CYAN
        self.grey = colorama.Fore.LIGHTBLACK_EX
        self.normal = colorama.Fore.RESET

    def stdout(self, msg, colour_code=colorama.Fore.RESET, verbosity=0, overwritable=False):
        # msg           String containing message for stdout.
        # color         stdout code (e.g. '\033[92m')
        # verbosity     0 = Always shown
        #               1 = -v flag
        #               2 = -vv flag
        if overwritable:
            line_end = "\r"
        else:
            line_end = "\n"

        if self.verbose_level >= verbosity:
            # Only colourise output if running in a real terminal.
            if sys.stdout.isatty():
                print(colour_code + msg + '\033[0m', end=line_end)
            else:
                print(msg)


def get_exception_as_string(e):
    """
    For when things go wrong, convert an exception object into a human-readable
    output that is normally displayed via the GUI.
    """
    return traceback.format_exc().replace("'", '’').replace('"', '’')


def is_exception_fault_by_app(e):
    """
    Parses the exception and return a boolean to indicate the crash was
    caused by this application. Usually when interfacing with a backend's library.
    This can help inform the user which project to consult.
    """
    if isinstance(e, BaseException):
        lines = traceback.format_exc().splitlines()
    else:
        lines = e.split("\n")
    files = []
    for line in lines:
        if line.strip().startswith("File"):
            files.append(line)

    # Last file path mentions our name, most likely our code.
    if files[-1].find("polychromatic") >= 0:
        return True

    # Last file is another library that caused the error.
    return False


def get_form_factor(_, form_factor_id):
    """
    Reads a string provided by a backend and returns data that is used to refer
    to the device ("form factor") throughout the application.

    Params:
        form_factor_id      (str)   One string from FORM_FACTORS.

    Returns:
        {id, icon, label}   A dictionary consisting of:
                                id          Form Factor ID
                                icon        Absolute path to icon
                                label       Human-readable name of form factor
    """
    if form_factor_id not in FORM_FACTORS:
        form_factor_id = "unrecognised"

    labels = {
        "accessory": _("USB Accessory"),
        "display": _("Display"),
        "fan": _("Fan"),
        "gpu": _("Graphics Card"),
        "headset": _("Headset"),
        "keyboard": _("Keyboard"),
        "keypad": _("Keypad"),
        "laptop": _("Laptop"),
        "mousemat": _("Mousemat"),
        "mouse": _("Mouse"),
        "ram": _("Memory"),
        "speaker": _("Speaker"),
        "stand": _("Stand"),
        "unrecognised": _("Unrecognised")
    }

    return {
        "id": form_factor_id,
        "icon": get_icon("devices", form_factor_id),
        "label": labels[form_factor_id]
    }


def get_green_shades(_):
    """
    Returns a custom colours.json for use with non-RGB keyboards,
    like the Razer BlackWidow Ultimate.
    """
    colours = []
    count = 0
    for shade in ["#00FF00", "#00E100", "#00C800", "#00AF00", "#009600", "#007D00", "#006400", "#004B00", "#003200"]:
        count += 1
        colours.append({
            "name": "{0} {1}".format(_("Green"), str(count)),
            "hex": shade
        })
    return colours


def get_default_tray_icon():
    """
    Determines which tray icon is best suited for the current desktop environment
    or theme.

    The path is intentionally relative as this is saved to preferences.json.
    A relative path will look up the icon in Polychromatic's data folders.
    """
    desktop_env = os.environ.get("XDG_CURRENT_DESKTOP")
    theme_env = os.environ.get("GTK_THEME")

    # Default icon
    icon_value = "img/tray/light/polychromatic.svg"

    # TODO: Detect GTK dark theme.

    if desktop_env:
        if desktop_env == "KDE":
            icon_value = "img/tray/light/breeze.svg"

    elif theme_env:
        # Unity/Ubuntu MATE
        if theme_env.startswith("Ambiant") or theme_env.startswith("Ambiance"):
            icon_value = "img/tray/light/humanity.svg"

    return icon_value


def get_tray_icon(dbg, icon_value):
    """
    Returns the full path to the icon to use with the tray applet.

    Params:
        dbg             (obj)   Application's "dbg" object
        icon_value      (str)   Preference value of ["tray"]["icon"]
    """
    # Check if the icon is absolute -> a custom icon
    if os.path.exists(icon_value):
        return icon_value

    # Check if the icon is relative -> a built-in icon
    icon_builtin = os.path.join(paths.data_dir, icon_value)
    if os.path.exists(icon_builtin):
        return icon_builtin

    dbg.stdout("Tray icon missing: {0}\nUsing fallback icon.".format(icon_value), dbg.warning)
    return get_icon("tray/light", "polychromatic")


def get_icon(folder, name):
    """
    Returns the absolute path to a Polychromatic provided icon from the data
    folder.

    Example:
        ("general", "battery-75") returns "/usr/share/polychromatic/img/general/battery-75.svg"

    Returns:
        (str)       Absolute path to icon
        None        Icon does not exist
    """
    for ext in [".svg", ".png"]:
        icon_path = os.path.join(paths.data_dir, "img", folder, name + ext)
        if os.path.exists(icon_path):
            return icon_path
    return None


def generate_colour_bitmap(dbg, colour_hex, size=22):
    """
    Generates a small bitmap of a colour and returns the path. Used for some
    UI controls that cannot use stylesheets.

    The file is cached to speed up future retrievals of the colour.
    """
    # FIXME: Refactoring required: Remove 'dbg'
    cache_name = hashlib.md5(str(colour_hex + str(size)).encode("utf-8")).hexdigest()
    cache_path = os.path.join(paths.assets_cache, cache_name + ".svg")
    cache_dir = os.path.dirname(cache_path)

    if not os.path.exists(cache_path):
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)

        with open(cache_path, "w") as f:
            f.writelines("<svg xmlns=\"http://www.w3.org/2000/svg\" version=\"1.1\" " + \
                "width=\"{0}\" height=\"{0}\"><rect width=\"{0}\" height=\"{0}\" ".format(str(size)) + \
                    "style=\"fill:{0}\"/></svg>".format(colour_hex))

    if not os.path.exists(cache_path):
        dbg.stdout("ERROR: Failed to generate colour SVG: " + colour_hex, dbg.error)
        return None

    return cache_path


def get_icon_styles(dbg, folder, name, normal_colour, disabled_colour, active_colour, selected_colour, secondary_active, secondary_inactive):
    """
    Returns a list of icon paths to SVG assets for use with buttons and other
    Qt widgets in this order: ["normal", "disabled", "active", "selected"]. If
    the icon is missing, then None is returned.

    Paramaters are for get_icon() and the hex values to use. The secondary colour
    is used to recolour icons with a dual tone.

    The file is cached to speed up future retrievals of the asset.
    """
    original_icon = get_icon(folder, name)
    icons = []

    if not original_icon:
        return None

    for colour in [normal_colour, disabled_colour, active_colour, selected_colour]:
        cache_name = hashlib.md5(str(folder + name + colour).encode("utf-8")).hexdigest()
        cache_path = os.path.join(paths.assets_cache, cache_name + ".svg")
        cache_dir = os.path.dirname(cache_path)

        if not os.path.exists(cache_path):
            dbg.stdout("Generating icon style: {0}/{1} ({2})".format(folder, name, colour), dbg.action, 1)

            with open(original_icon, "r") as f:
                data = f.readlines()

            newdata = []
            for line in data:
                secondary_colour = secondary_active if colour in [active_colour, selected_colour] else secondary_inactive
                newdata.append(line.replace("#00FF00", colour).replace("#00ff00", colour).replace("#008000", secondary_colour))

            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir)

            with open(cache_path, "w") as f:
                f.writelines(newdata)

        icons.append(cache_path)

    return icons


def get_full_path_for_save_data_icon(icon_path):
    """
    Returns the full path to an icon specified in the save data for the UI to use.

    Polychromatic stores icon paths relatively. These paths are to be checked
    in this priority:
        - Custom Icons (~/.config/.../custom_icons/example.png)
        - Emblems/Tray (/usr/share/.../img/emblems/example.svg)

    If the icon no longer exists, a fallback will be provided.
    """
    possible_paths = [
        os.path.join(paths.custom_icons, icon_path),
        os.path.join(paths.data_dir, icon_path)
    ]

    if os.path.exists(icon_path):
        # Icon is already absolute!
        return icon_path

    # Try relative paths
    for path in possible_paths:
        if os.path.exists(path):
            return path

    return get_icon("devices", "unrecognised")


def execute_polychromatic_component(dbg, component, controller_open=None):
    """
    Starts a Polychromatic application relative to its location or system-wide
    if installed.

    Params:
        component           e.g. "controller" would run "polychromatic-controller"
        controller_open     (Optional - Controller only) Opens a specific tab/feature.
    """
    data_dir = paths.data_dir
    exec_name = "polychromatic-" + component

    possible_paths = [
        # Relative copy #1
        os.path.join(data_dir, "../" + exec_name),

        # Relative copy #2
        os.path.join(os.path.dirname(__file__), exec_name),

        # Relative system-wide (e.g. /usr/share/polychromatic)
        os.path.join(data_dir, "../../bin/" + exec_name),

        # Relative system-wide (e.g. /usr/share/polychromatic)
        os.path.join(os.path.dirname(__file__), "../" + exec_name),

        # Absolute system-wide
        "/usr/bin/" + exec_name
    ]

    for bin_path in possible_paths:
        args = [bin_path]

        if controller_open:
            args.append("--open")
            args.append(controller_open)

        if os.path.exists(bin_path):
            dbg.stdout("Executing: " + os.path.realpath(" ".join(args)), dbg.debug, 1)
            try:
                subprocess.Popen(args)
            except Exception:
                pass
            return True
    return False


def run_thread(target_function, args=()):
    """
    Executes a function that will run outside the main thread.

    target_function     Function to execute.
    args                (Optional) A tuple containing arguments.
    """
    thread = Thread(target=target_function, args=args)
    thread.daemon = True
    thread.start()
    return thread


def rgb_to_hex(rgb_list):
    """
    Converts [R,G,B] list to #RRGGBB string.
    Polychromatic stores and processes colours as hex values.
    """
    return "#{0:02X}{1:02X}{2:02X}".format(*rgb_list)


def hex_to_rgb(hex_string):
    """
    Converts "#RRGGBB" string to [R,G,B] list.
    Some backends/logic may expect colours to be individual RGB values.
    """
    hex_string = hex_string.lstrip("#")
    return list(int(hex_string[i:i+2], 16) for i in (0, 2 ,4))


def validate_hex(value):
    """
    Validates a colour hex value and returns whether it's valid. Must be a
    string, starting with # and consisting of 6 hex digits (0-F).
    """
    if not value.startswith("#") or not len(value) == 7:
        return False

    try:
        int(value[1:7], 16)
    except ValueError:
        return False

    return True


def get_plural(integer, non_plural, plural):
    """
    Returns the correct plural or non-plural spelling based on an integer.
    """
    if integer == 1:
        return non_plural
    else:
        return plural


def get_versions(base_version):
    """
    When running from a Git repository, return the development revision and
    Git commit "revision" as a tuple. Otherwise just the release version.

    This is intended to make debugging easier.
    """
    version = base_version
    git_commit = None
    py_version = "{0}.{1}.{2}".format(sys.version_info.major, sys.version_info.minor, sys.version_info.micro)

    if os.path.exists(os.path.join(os.path.dirname(__file__), "..", ".git")):
        import subprocess

        old_path = os.getcwd()
        os.chdir(os.path.dirname(__file__))

        try:
            git_commit = subprocess.check_output(["git", "rev-parse", "HEAD"]).strip().decode("UTF-8")

            try:
                version = subprocess.check_output(["git", "describe", "--tags"]).strip().decode("UTF-8")[1:]
            except subprocess.CalledProcessError:
                # Git may throw error "no names found" if repository was shallow cloned or has no tags
                version = "{0}-git-{1}".format(base_version, git_commit[:7])

        except subprocess.CalledProcessError:
            # "git" failed to run, this could happen due to "untrusted" repositories
            version = base_version + "~????????-git"

        except FileNotFoundError:
            # "git" is not installed despite being a git repository
            version = base_version + "-git"

        os.chdir(old_path)

    # Production "installed" version
    return (version, git_commit, py_version)



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2020-2024 Luke Horwell <code@horwell.me>
"""
This module contains the frontend for the troubleshooter.
"""

import os
import time
import subprocess
import webbrowser

from .. import common
from .. import preferences
from .. import procpid
from . import shared
from . import procviewer

from PyQt5.QtCore import Qt, QThread, pyqtSignal, pyqtSlot, QObject
from PyQt5.QtGui import QIcon, QBrush
from PyQt5.QtWidgets import QWidget, QLabel, QDialog, QPushButton, \
                            QTreeWidget, QTreeWidgetItem, QProgressBar, \
                            QApplication


class TroubleshooterGUI(QObject):
    """
    Runs the troubleshooter and shows the results in a dialog box.
    """
    signal_progress_bar_advance = pyqtSignal()
    signal_progress_bar_set_max = pyqtSignal(int)

    def __init__(self, appdata, backend, backend_human_name):
        # Required for slot/signals to work
        super().__init__()

        self.appdata = appdata
        self.backend = backend
        self.human_name = backend_human_name
        self._ = appdata._
        self.dbg = self.appdata.dbg
        self.widgets = shared.PolychromaticWidgets(appdata)

        self.result_window = shared.get_ui_widget(self.appdata, "troubleshooter", QDialog)
        self.result_label = self.result_window.findChild(QLabel, "Title")
        self.result_tree = self.result_window.findChild(QTreeWidget, "Results")
        self.result_tree_root = self.result_tree.invisibleRootItem()
        self.result_copy_btn = self.result_window.findChild(QPushButton, "CopyToClipboard")

        if not self.appdata.system_qt_theme:
            self.result_window.findChild(QPushButton, "Close").setIcon(self.widgets.get_icon_qt("general", "close"))

        self.loading = shared.get_ui_widget(self.appdata, "loading", QDialog)
        self.loading_label = self.loading.findChild(QLabel, "Label")
        self.loading_progress_bar = self.loading.findChild(QProgressBar, "ProgressBar")

        self.signal_progress_bar_advance.connect(self.progress_bar_advance)
        self.signal_progress_bar_set_max.connect(self.progress_bar_set_max)

        self.thread = self.TroubleshootThread()
        self.thread.parent = self
        self.thread.result = []
        self.thread.finished.connect(self.finished)

        self.start()

    def progress_bar_advance(self):
        self.loading_progress_bar.setValue(self.loading_progress_bar.value() + 1)

    def progress_bar_set_max(self, max_value):
        self.loading_progress_bar.setRange(0, 0)
        self.loading_progress_bar.setMaximum(max_value)

    class TroubleshootThread(QThread):
        """
        Troubleshooting runs in a separate thread to prevent UI lock up, in case
        the code takes longer to run on some systems.
        """
        def run(self):
            _ = self.parent._

            while not self.parent.appdata.ready:
                self.parent.loading_label.setText(_("Waiting for backends to be ready..."))
                time.sleep(0.1)

            self.parent.loading_label.setText(_("Running troubleshooter..."))
            self.result = self.parent.appdata.middleman.troubleshoot(self.parent.backend, _, self.parent.signal_progress_bar_set_max.emit, self.parent.signal_progress_bar_advance.emit)

    def start(self):
        self.dbg.stdout("Running troubleshooter backend: {0}".format(self.backend), self.dbg.action, 1)
        self.loading_label.setText(self._("Running troubleshooter..."))
        self.loading.setWindowTitle(self._("Troubleshooting..."))
        self.loading.setWindowFlag(Qt.WindowMinimizeButtonHint, False)
        self.loading.setWindowFlag(Qt.WindowMaximizeButtonHint, False)
        self.loading.open()

        self.thread.start()

    def finished(self):
        results = self.thread.result
        all_passed = True
        self.loading.close()

        if type(results) == None:
            return self.widgets.open_dialog(self.widgets.dialog_warning,
                                            self._("Troubleshooting Failed"),
                                            self._("The troubleshooter for this backend is not available for this operating system."))

        elif type(results) == str:
            return self.widgets.open_dialog(self.widgets.dialog_error,
                                            self._("Troubleshooting Failed"),
                                            self._("An exception was thrown while running the troubleshooter. This is probably a bug."),
                                            details=results)

        for result in results:
            item = QTreeWidgetItem()

            # "passed" must be either: True (Passed); False (Failed); None (Unknown)
            if result["passed"] is True:
                item.setText(0, self._("Passed"))
                item.setIcon(0, QIcon(common.get_icon("general", "success")))
            elif result["passed"] is False:
                item.setText(0, self._("Failed"))
                item.setIcon(0, QIcon(common.get_icon("general", "serious")))
            else:
                item.setText(0, self._("Unknown"))
                item.setIcon(0, QIcon(common.get_icon("general", "unknown")))

            item.setText(1, result["test_name"])
            item.copyable = False

            # Provide suggestions on failures
            if not result["passed"]:
                all_passed = False

                suggestions = result["suggestions"]
                for line in suggestions:
                    subitem = QTreeWidgetItem()
                    if line.startswith("$"):
                        text = line
                        subitem.setBackground(1, QBrush(Qt.black))
                        subitem.setForeground(1, QBrush(Qt.green))
                        subitem.copyable = True
                    else:
                        text = "• " + line
                        subitem.setDisabled(True)
                    subitem.setText(1, text)
                    subitem.setToolTip(1, line)
                    item.addChild(subitem)
            self.result_tree_root.addChild(item)

        self.result_tree.expandAll()
        self.result_window.findChild(QPushButton, "Close").clicked.connect(self._close_troubleshooter)
        self.result_tree.itemSelectionChanged.connect(self._item_changed)
        self.result_copy_btn.setHidden(True)
        self.result_copy_btn.clicked.connect(self._copy_to_clipboard)
        self.result_window.setWindowTitle(self._("Troubleshooter for []").replace("[]", self.human_name))
        if all_passed:
            self.result_label.setText(self._("Everything appears to be in working order!"))
        self.dbg.stdout("Troubleshooting completed!", self.dbg.success, 1)
        self.result_window.open()

    def _item_changed(self):
        """
        User clicks on one of the results. Shows/hides the "Copy" button.
        """
        selected = self.result_tree.selectedItems()[0]
        self.result_copy_btn.setHidden(selected.copyable == False)

    def _copy_to_clipboard(self):
        """
        Commands can be copied to clipboard for convenience
        """
        selected_text = self.result_tree.selectedItems()[0].text(1).replace("$ ", "")
        QApplication.clipboard().setText(selected_text)

    def _close_troubleshooter(self):
        """
        """
        self.result_window.close()
        self.result_window.deleteLater()
        self.loading.deleteLater()



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2020-2024 Luke Horwell <code@horwell.me>
"""
This module controls the 'Effects' tab of the Controller GUI.
"""

from .. import common
from .. import effects
from .. import locales
from .. import procpid
from .. import preferences as pref
from . import shared
from . import editor

import os

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon, QPixmap
from PyQt5.QtWidgets import QWidget, QPushButton, QToolButton, QMessageBox, \
                            QListWidget, QTreeWidget, QLabel, QComboBox, \
                            QTreeWidgetItem, QMenu, QDialog, QDialogButtonBox, \
                            QButtonGroup, QLineEdit, QTextEdit, QCheckBox, \
                            QGroupBox, QRadioButton, QMainWindow


class EffectsTab(shared.CommonFileTab):
    """
    Allows the user to manage software effects for lighting up their devices.
    """
    def __init__(self, appdata):
        super().__init__(appdata, effects.EffectFileManagement, "EffectContents", "EffectSidebarTree")
        self.feature = "effects"
        self.tab_title = self._("Effects")

        # Populate tasks
        self.tasks = {
            "new": self.new_file,
            "import": self.import_effect
        }

        self._add_tree_item(self.TasksBranch, self._("New Effect"), common.get_icon("general", "new"), "tasks", "new")
        # FIXME: Not yet implemented: Import Effect
        #self._add_tree_item(self.TasksBranch, self._("Import Effect"), common.get_icon("general", "import"), "tasks", "import")

        # Keep track of editor windows so garbage collection doesn't destroy them
        self.editors = {}

    def show_no_file_screen(self, message_id):
        """
        Effect cannot be opened for viewing - inform the user.

        Params:
            message_id  (int)   0   Empty file list
                                1   File corrupt
        """
        layout = self.Contents.layout()

        titles = {
            0: self._("It's empty here!"),
            1: self._("Error loading effect")
        }

        subtitles = {
            0: self._("Your own software effects appear here. Try creating one!"),
            1: self._("There's invalid data in this file. The file might be corrupt.")
        }

        icons = {
            0: None,
            1: None
        }

        buttons = {
            0: [
                {
                    "label": self._("New Effect"),
                    "icon_folder": "general",
                    "icon_name": "new",
                    "action": self.new_file
                }
                # FIXME: Not yet implemented: Import Effect
                #{
                    #"label": self._("Import Effect"),
                    #"icon_folder": "general",
                    #"icon_name": "import",
                    #"action": self.import_effect
                #}
            ],
            1: []
        }

        shared.clear_layout(layout)
        self.widgets.populate_empty_state(layout, icons[message_id], titles[message_id], subtitles[message_id], buttons[message_id])

    def _check_for_device_new_file(self):
        """
        Makes sure there is a device connected required to create an effect.

        The current codebase doesn't provide a mechanism to list all compatible
        hardware yet, since effects need to store their rows/cols values.
        """
        if len(self.middleman.get_devices()) == 0:
            self.widgets.open_dialog(self.widgets.dialog_error,
                                     self._("New Effect"),
                                     self._("No devices found. New effects require a compatible device to be present."),
                                     info_text=self._("Make sure the device is connected and the backend is working."))
            return False
        return True

    def new_file(self):
        """
        Prompts the user which type of effect to create.
        """
        if not self._check_for_device_new_file():
            return

        dialog = shared.get_ui_widget(self.appdata, "new-effect", QDialog)

        # TODO: Not all effect types are implemented, create sequence one
        self.new_file_stage_2(dialog, effects.TYPE_SEQUENCE)
        return

        btn_layered = dialog.findChild(QToolButton, "NewLayered")
        btn_scripted = dialog.findChild(QToolButton, "NewScripted")
        btn_sequence = dialog.findChild(QToolButton, "NewSequence")
        hover_tooltip = dialog.findChild(QLabel, "HoverText")
        dialog_buttons = dialog.findChild(QDialogButtonBox, "DialogButtons")

        # Load icons
        # FIXME: QToolButton() icon not behaving like QPushButton() when button is focused?
        btn_layered.setIcon(self.widgets.get_icon_qt("effects", "layered"))
        btn_scripted.setIcon(self.widgets.get_icon_qt("effects", "scripted"))
        btn_sequence.setIcon(self.widgets.get_icon_qt("effects", "sequence"))

        # Hovering over a button will show the tooltip as a label too
        def _hover_button(event, button):
            hover_tooltip.setText(button.toolTip())
            button.repaint()

        btn_layered.enterEvent = lambda event: _hover_button(event, btn_layered)
        btn_scripted.enterEvent = lambda event: _hover_button(event, btn_scripted)
        btn_sequence.enterEvent = lambda event: _hover_button(event, btn_sequence)

        # For keyboard navigation
        btn_layered.focusInEvent = lambda event: _hover_button(event, btn_layered)
        btn_scripted.focusInEvent = lambda event: _hover_button(event, btn_scripted)
        btn_sequence.focusInEvent = lambda event: _hover_button(event, btn_sequence)

        # Prepare button events
        def _close_dialog():
            dialog.reject()
        dialog_buttons.rejected.connect(_close_dialog)

        btn_layered.clicked.connect(lambda: self.new_file_stage_2(dialog, effects.TYPE_LAYERED))
        btn_scripted.clicked.connect(lambda: self.new_file_stage_2(dialog, effects.TYPE_SCRIPTED))
        btn_sequence.clicked.connect(lambda: self.new_file_stage_2(dialog, effects.TYPE_SEQUENCE))

        # FIXME: Not yet implemented editors
        btn_layered.setEnabled(False)
        btn_scripted.setEnabled(False)

        dialog.open()

    def new_file_stage_2(self, old_dialog, effect_type):
        """
        Open the metadata window for the user to enter details for their new creation.
        On successful initialisation, the effect editor will open.

        Params:
            old_dialog      QDialog() from new_file()
            effect_type     effects.TYPE_* reference
        """
        old_dialog.accept()
        data = self.fileman.init_data("", effect_type)

        # Set default icon to what the effect is
        if effect_type == effects.TYPE_LAYERED:
            data["icon"] = "img/effects/layered.svg"
        elif effect_type == effects.TYPE_SCRIPTED:
            data["icon"] = "img/effects/scripted.svg"
        elif effect_type == effects.TYPE_SEQUENCE:
            data["icon"] = "img/effects/sequence.svg"

        def _save_metadata(newdata):
            """
            After closing the metadata editor, save the data to disk and open
            the editor according to the effect.
            """
            success, path = self.fileman.save_item(newdata)

            if not success:
                self.show_error_message(path, self._show_file_error())
                self.new_file_stage_2(metadata_editor, effect_type)
                return

            # Open the editor
            self.current_file_path = path
            self.current_file_data = newdata
            self.edit_file()

            # Refresh the UI for when the user returns
            self.set_tab()
            self.open_file(path)

        # Backends not ready - cannot show device list
        if not self.appdata.ready:
            self.dbg.stdout("Backends not ready! Device list will not be available for new effect.", self.dbg.warning, 1)
            action = ""

            def do_ignore():
                # Let the metadata editor handle no devices
                pass

            def do_retry():
                nonlocal action
                action = "retry"

            def do_cancel():
                nonlocal action
                action = "cancel"

            self.widgets.open_dialog(self.widgets.dialog_warning,
                                     self._("Backend Not Ready"),
                                     self._("The application is still waiting for the backends to finish initialising so you can pick a device to map your new effect."),
                                     info_text=self._("This might be ready in a few moments."),
                                     buttons=[QMessageBox.Ignore, QMessageBox.Retry],
                                     actions={
                                         QMessageBox.Ignore: do_ignore,
                                         QMessageBox.Retry: do_retry
                                     })

            if action == "cancel":
                return
            elif action == "retry":
                return self.new_file_stage_2(old_dialog, effect_type)

        metadata_editor = EffectMetadataEditor(self.appdata, data, _save_metadata)

    def open_file(self, effect_path):
        """
        Opens the specified file in the interface, providing an overview of the
        effect and buttons to modify the file.
        """
        self.current_file_path = effect_path
        self.current_file_data = self.fileman.get_item(effect_path)
        data = self.current_file_data

        if type(data) == int:
            self.show_error_message(effect_path, data)
            self.show_no_file_screen(1)
            return

        layout = self.Contents.layout()
        shared.clear_layout(layout)

        # Summary
        icon_path = data["parsed"]["icon"]
        buttons = [
            {
                "id": "play",
                "icon": self.widgets.get_icon_qt("effects", "play"),
                "label": self._("Play"),
                "disabled": False,
                "action": self.play_effect
            },
            {
                "id": "edit",
                "icon": self.widgets.get_icon_qt("general", "edit"),
                "label": self._("Edit"),
                "disabled": False,
                "action": self.edit_file
            },
            {
                "id": "clone",
                "icon": self.widgets.get_icon_qt("general", "clone"),
                "label": self._("Clone"),
                "disabled": False,
                "action": self.clone_file
            },
            {
                "id": "delete",
                "icon": self.widgets.get_icon_qt("general", "delete"),
                "label": self._("Delete"),
                "disabled": False,
                "action": self.delete_file
            }
        ]

        # Populate indicators
        indicators = []

        # -- Author
        if data["author"]:
            indicators.append({"icon": common.get_icon("effects", "author"), "label": data["author"]})

        # -- Effect Type
        effect_type = data["type"]
        effect_type_name = {
            effects.TYPE_LAYERED: self._("Layered"),
            effects.TYPE_SCRIPTED: self._("Script"),
            effects.TYPE_SEQUENCE: self._("Sequence"),
        }
        effect_type_icon = {
            effects.TYPE_LAYERED: "layered",
            effects.TYPE_SCRIPTED: "scripted-small",
            effects.TYPE_SEQUENCE: "sequence",
        }
        indicators.append({"icon": common.get_icon("effects", effect_type_icon[effect_type]), "label": effect_type_name[effect_type]})

        # -- Device
        if data["map_device"]:
            device_icon = common.get_icon("devices", data["map_device_icon"])
            if not device_icon:
                device_icon = common.get_icon("devices", "accessory")
            indicators.append({"icon": device_icon, "label": data["map_device"]})

        # Create the summary widget
        summary = self.widgets.create_summary_widget(icon_path, data["parsed"]["name"], indicators, buttons)
        summary.setMaximumHeight(170)
        layout.addWidget(summary)

        # Show the file's summary text (if any)
        summary_text = data["summary"]
        if len(summary_text) > 0:
            summary_label = QLabel()
            summary_label.setText(summary_text)
            summary_label.setWordWrap(True)
            layout.addWidget(self.widgets.create_row_widget(self._("Summary"), [summary_label]))

        # Show other actions for this effect
        def _edit_metadata():
            # To prevent data loss, don't edit metadata here if the editor is open
            if effect_path in self.editors.keys():
                if self.editors[effect_path].alive:
                    self.widgets.open_dialog(self.widgets.dialog_warning,
                                             self._("File In Use"),
                                             self._("This file is being edited in another window. Please switch to that window to make changes."))
                    return

            def _metadata_changed(newdata):
                result, new_path = self.fileman.save_item(newdata, effect_path)
                self.open_file(new_path)
                if len(self.Sidebar.selectedItems()) > 0:
                    self.Sidebar.selectedItems()[0].setText(0, newdata["name"])
                    self.Sidebar.selectedItems()[0].setIcon(0, QIcon(common.get_full_path_for_save_data_icon(newdata["icon"])))

            self.metadata_editor = EffectMetadataEditor(self.appdata, self.current_file_data, _metadata_changed)

        btn_edit_meta = QPushButton()
        btn_edit_meta.setText(self._("Edit Metadata"))
        btn_edit_meta.setIcon(self.widgets.get_icon_qt("general", "properties"))
        btn_edit_meta.clicked.connect(_edit_metadata)
        widget_edit_meta = self.create_widget_wrapper_for_control([btn_edit_meta])

        other_actions = self.widgets.create_row_widget(self._("Other Actions"), [widget_edit_meta], vertical=True)
        other_actions.findChild(QLabel).setContentsMargins(0, 4, 0, 4)
        layout.addWidget(other_actions)
        layout.addStretch()

    def edit_file(self):
        """
        Open the editor for the currently selected effect. Make sure only one
        editor is open per file.
        """
        effect_path = self.current_file_path
        effect_type = self.current_file_data["type"]

        if effect_path in self.editors.keys():
            if self.editors[effect_path].alive:
                self.widgets.open_dialog(self.widgets.dialog_warning,
                                         self._("File In Use"),
                                         self._("This file is being edited in another window. Please close that editor first."))
                return

        if effect_type in [effects.TYPE_LAYERED, effects.TYPE_SEQUENCE]:
            self.editors[effect_path] = editor.VisualEffectEditor(self.appdata, self.fileman, effect_path)

    def import_effect(self):
        """
        Allows the user to create effects from other media, such as videos or images.
        """
        print("stub:effects.import_effect")
        pass

    def play_effect(self, device_name=None):
        """
        Play the currently selected effect.

        Optionally, for scripted effects, they could be compatible to run on one
        or more hardware.
        """
        effect_type = self.current_file_data["type"]
        if not device_name:
            device_name = self.current_file_data["map_device"]

        procmgr = procpid.ProcessManager("helper")
        procmgr.start_component(["--run-fx", self.current_file_path, "--device-name", device_name])


class EffectMetadataEditor(shared.TabData):
    """
    A dialog for modifying the metadata of an effect. This works with data
    that is currently residing in memory.
    """
    def __init__(self, appdata, data, callback_fn):
        """
        Prepare and show the window for editing the metadata of a saved (or
        new) effect.

        Params:
            appdata         ApplicationData() object
            data            Effect data
            callback_fn     Run function after save. The parameter will be the new data.
        """
        super().__init__(appdata)
        self.data = data
        self.effect_type = data["type"]
        self.callback_fn = callback_fn
        self.mapping_graphics = effects.DeviceMapGraphics(appdata)
        self.graphic_list = self.mapping_graphics.get_graphic_list()

        # Session: Remember device attributes: icon, cols, rows
        self.device_info = {}

        # Dialog & Controls
        self.dialog = shared.get_ui_widget(self.appdata, "effect-metadata-editor", QDialog)
        self.buttons = self.dialog.findChild(QDialogButtonBox, "DialogButtons")

        # -- Left: Details
        self.name = self.dialog.findChild(QLineEdit, "EffectName")
        self.author = self.dialog.findChild(QLineEdit, "Author")
        self.icon = self.dialog.findChild(QLabel, "EffectIconPlaceholder")
        self.summary = self.dialog.findChild(QTextEdit, "Summary")

        # -- Right: Hardware
        self.map_device_combo = self.dialog.findChild(QComboBox, "MapDeviceCombo")
        self.map_graphic_grid = self.dialog.findChild(QRadioButton, "MapGraphicToGrid")
        self.map_graphic_svg = self.dialog.findChild(QRadioButton, "MapGraphicToSVG")
        self.map_graphic_list = self.dialog.findChild(QComboBox, "MapGraphicList")
        self.map_preview = self.dialog.findChild(QLabel, "MappingPreview")
        self.dimensions_device = self.dialog.findChild(QLabel, "DeviceDimensions")

        self._init_controls()

    def _init_controls(self):
        """
        Populates and set up the UI for this window.
        """
        # Prepare icon picker
        def _update_icon(new_icon):
            self.data["icon"] = new_icon

        picker = self.widgets.create_icon_picker_control(_update_icon, self.data["icon"], self._("Choose Effect Icon"))
        self.icon.parentWidget().layout().replaceWidget(self.icon, picker)
        self.icon.deleteLater()
        self.icon = picker

        # Scripted effects do not configure mapping
        if self.effect_type == effects.TYPE_SCRIPTED:
            self.dialog.findChild(QGroupBox, "MappingGroup").setHidden(True)
            self.dialog.adjustSize()

        # Populate metadata fields
        self.name.setText(self.data["name"])
        self.author.setText(self.data["author"])
        self.summary.setText(self.data["summary"])

        # Mapping Configuration (non-scripted effects only)
        if not self.effect_type == effects.TYPE_SCRIPTED:
            # Populate device list
            if not self._populate_devices():
                return

            # Select appropriate mapping mode
            if self.data["map_graphic"]:
                self.map_graphic_svg.setChecked(True)
            else:
                self.map_graphic_grid.setChecked(True)
            self._update_graphic_preview()

        # Name is a required field
        name_label = self.dialog.findChild(QLabel, "EffectNameLabel")
        name_label.setText(name_label.text() + "*")

        # Connect signals
        self.dialog.accepted.connect(self._save_changes)
        self.name.textChanged.connect(self._validate_fields)
        self.map_device_combo.currentIndexChanged.connect(self._validate_fields)
        self.map_device_combo.currentIndexChanged.connect(self._device_updated)
        self.map_graphic_grid.toggled.connect(self._set_map_grid)
        self.map_graphic_svg.toggled.connect(self._set_mode_graphic)
        self.map_graphic_list.currentIndexChanged.connect(self._update_graphic_preview)

        # Showtime!
        self._validate_fields()
        self.dialog.open()

    def _populate_devices(self):
        """
        Populate the list of compatible devices to create a software effect for.

        For effects already created, only identical devices (cols/rows) may be
        interchanged (although there is no guarantee that the matrixes map 1:1)
        """
        is_new_file = len(self.data["name"]) == 0
        found_device_name = False

        for device in self.middleman.get_devices():
            if not device or not device.matrix:
                continue

            self.map_device_combo.addItem(QIcon(device.form_factor["icon"]), device.name)
            index = self.map_device_combo.count() - 1

            # Store device data for later
            self.device_info[index] = {}
            self.device_info[index]["icon"] = device.form_factor["id"]
            self.device_info[index]["cols"] = device.matrix.cols
            self.device_info[index]["rows"] = device.matrix.rows

            if self.data["map_device"] == device.name:
                found_device_name = True
                self.map_device_combo.setCurrentIndex(index)

            # Disable option for existing effects and does not match another device
            if not is_new_file:
                if not self.data["map_cols"] == device.matrix.cols or \
                    not self.data["map_rows"] == device.matrix.rows:
                        self.map_device_combo.model().item(index).setEnabled(False)

        # New effect? Select the first device.
        if not self.data["map_device"]:
            self.dialog.setWindowTitle(self._("New Effect"))
            self.map_device_combo.setCurrentIndex(0)

        # Add a placeholder for device if it is not present
        elif not found_device_name and not is_new_file:
            self.map_device_combo.addItem(QIcon(common.get_icon("devices", "unrecognised")), self.data["map_device"])
            index = self.map_device_combo.count() - 1
            self.map_device_combo.setCurrentIndex(index)
            self.device_info[index] = {}
            self.device_info[index]["icon"] = self.data["map_device_icon"]
            self.device_info[index]["cols"] = self.data["map_cols"]
            self.device_info[index]["rows"] = self.data["map_rows"]

        # No devices to map
        if self.map_device_combo.count() == 0:
            self.widgets.open_dialog(self.widgets.dialog_error,
                                     self._("New Effect"),
                                     self._("No compatible devices found. Only individually addressable LED capable hardware can be used for effects."))
            return False

        # Populate graphic list
        self._device_updated()
        return True

    def _validate_fields(self):
        """
        Ensure the fields have sufficient data before saving is allowed.
        """
        conditions = [
            # Must have a name
            len(self.name.text()) > 0
        ]

        self.buttons.button(QDialogButtonBox.Ok).setEnabled(all(x is True for x in conditions))

    def _device_updated(self):
        """
        Mapping device changed. Update the graphic list to ensure only compatible
        graphics are displayed.

        If the user is editing an existing effect and chooses a device with less
        columns/rows then previously, let the user know truncating might occur.
        """
        is_new_file = len(self.data["name"]) == 0
        device_name = self.map_device_combo.currentText()
        device_info = self.device_info[self.map_device_combo.currentIndex()]
        device_rows = device_info["rows"]
        device_cols = device_info["cols"]

        dimensions = common.get_plural(device_rows, self._("1 row"), self._("2 rows").replace("2", str(device_rows)))
        dimensions += ", " + common.get_plural(device_cols, self._("1 column"), self._("2 columns").replace("2", str(device_cols)))
        self.dimensions_device.setText(dimensions)

        # Filter graphics so only compatible matrix dimensions are shown
        self.map_graphic_list.clear()
        found_graphic_index = None
        for name in self.graphic_list.keys():
            graphic = self.graphic_list[name]

            if device_rows != graphic["rows"] or device_cols != graphic["cols"]:
                continue

            new_index = self.map_graphic_list.count()
            self.map_graphic_list.addItem(name)

            if self.data["map_graphic"] == graphic["filename"]:
                self.map_graphic_list.setCurrentIndex(new_index)
                found_graphic_index = new_index

        # If there are no graphics, only grid can be selected
        if self.map_graphic_list.count() > 0:
            self.map_graphic_svg.setEnabled(True)
            self.map_graphic_list.setEnabled(True)
            self._auto_detect_suitable_graphic()
        else:
            self.map_graphic_svg.setEnabled(False)
            self.map_graphic_list.setEnabled(False)
            self.map_graphic_grid.setChecked(True)

        # For new effects, auto select the first graphic
        if is_new_file:
            self.map_graphic_svg.setChecked(True if self.map_graphic_list.count() > 0 else False)
            self._auto_detect_suitable_graphic()

        # For existing effects, (re)select the correct graphic in list
        if self.data["map_graphic"] and found_graphic_index:
            self.map_graphic_list.setCurrentIndex(found_graphic_index)

    def _auto_detect_suitable_graphic(self):
        """
        When loading a new device and the 'graphic' option is set, determine
        the best graphic (as there could be multiple with the same cols/rows)
        """
        self.map_graphic_list.setCurrentIndex(0)
        current_locale = self.i18n.get_current_locale()

        # Pick graphic closely matching device name and locale (if applicable)
        device_name = self.map_device_combo.currentText()
        for name in self.graphic_list.keys():
            graphic = self.graphic_list[name]
            locale = graphic["locale"]
            if name.find(device_name) != -1 and locale == current_locale:
                self.map_graphic_list.setCurrentText(name)

            if name.find(device_name) != -1 and not locale:
                self.map_graphic_list.setCurrentText(name)

    def _set_map_grid(self):
        """
        User selects the "Grid" radio button.
        """
        self.map_graphic_list.setEnabled(False)
        self.map_graphic_list.setCurrentIndex(-1)
        self._update_graphic_preview()

    def _set_mode_graphic(self):
        """
        User selects the "Graphic" radio button.
        """
        self.map_graphic_list.setEnabled(True)
        self._auto_detect_suitable_graphic()
        self._update_graphic_preview()

    def _update_graphic_preview(self):
        """
        The graphic options have changed, refresh the preview.
        """
        device_map = effects.DeviceMapGraphics(self.appdata)
        device_info = self.device_info[self.map_device_combo.currentIndex()]
        cols = device_info["cols"]
        rows = device_info["rows"]
        svg_path = None

        # -- Grid
        if self.map_graphic_grid.isChecked():
            svg_path = device_map.get_grid_path(cols, rows)

        # -- Graphic
        elif self.map_graphic_list.currentText() == "":
            return

        elif self.map_graphic_svg.isChecked():
            graphic = self.graphic_list[self.map_graphic_list.currentText()]
            filename = graphic["filename"]
            svg_path = device_map.get_graphic_path(filename)

        # Load SVG into view
        if svg_path:
            shared.set_pixmap_for_label(self.map_preview, svg_path, 256)

    def _save_changes(self):
        """
        Commit the changes and pass them to the callback function.
        """
        newdata = self.data.copy()

        # Metadata fields
        newdata["name"] = self.name.text()
        newdata["author"] = self.author.text()
        newdata["summary"] = self.summary.toPlainText()

        # Device mapping information
        if not newdata["type"] == effects.TYPE_SCRIPTED:
            device_name = self.map_device_combo.currentText()
            device_info = self.device_info[self.map_device_combo.currentIndex()]
            graphic_name = self.map_graphic_list.currentText()

            newdata["map_device"] = device_name
            newdata["map_device_icon"] = device_info["icon"]
            newdata["map_cols"] = device_info["cols"]
            newdata["map_rows"] = device_info["rows"]
            newdata["map_graphic"] = ""
            if graphic_name:
                newdata["map_graphic"] = self.graphic_list[graphic_name]["filename"]

        self.callback_fn(newdata)



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2021-2024 Luke Horwell <code@horwell.me>
"""
This module controls the 'Background Tasks' window accessible from the menu bar.
"""

from ..base import PolychromaticBase
from .. import common
from .. import procpid
from . import shared

import glob
import os
import time
import subprocess

from PyQt5.QtCore import Qt, QThread
from PyQt5.QtGui import QIcon, QPixmap
from PyQt5.QtWidgets import QWidget, QDialog, QPushButton, QTreeWidget, \
                            QLabel, QTreeWidgetItem


class ProcessViewer(PolychromaticBase):
    """
    The process viewer provides details into the procpid module implementation.
    Allows the user to view processes spawned by Polychromatic, stop them or
    restart them for diagnosis purposes.
    """
    def __init__(self, appdata):
        self.appdata = appdata
        self.widgets = shared.PolychromaticWidgets(appdata)

        # Session
        self.dialog_is_open = True

        # UI Controls
        self.dialog = shared.get_ui_widget(appdata, "procviewer", q_toplevel=QDialog)
        self.tree = self.dialog.findChild(QTreeWidget, "TasksTree")
        self.close = self.dialog.findChild(QPushButton, "Close")
        self.btn_refresh = self.dialog.findChild(QPushButton, "TasksRefresh")
        self.btn_reload = self.dialog.findChild(QPushButton, "TasksReloadAll")
        self.btn_stop = self.dialog.findChild(QPushButton, "TasksStopTask")

        # Set Dialog Button Icons
        if not self.appdata.system_qt_theme:
            self.close.setIcon(self.widgets.get_icon_qt("general", "close"))
            self.btn_refresh.setIcon(self.widgets.get_icon_qt("general", "refresh"))
            self.btn_reload.setIcon(self.widgets.get_icon_qt("general", "reset"))
            self.btn_stop.setIcon(self.widgets.get_icon_qt("general", "cancel"))

        # Connect signals when interacting with UI controls
        self.close.clicked.connect(self._close)
        self.tree.itemActivated.connect(self._change_item)
        self.tree.itemClicked.connect(self._change_item)
        self.btn_refresh.clicked.connect(self._refresh_list)
        self.btn_reload.clicked.connect(self._reload_all)
        self.btn_stop.clicked.connect(self._stop_task)

        # Showtime!
        self._refresh_list()
        self.tree.setColumnWidth(0, 250)
        self.tree.setColumnWidth(1, 300)
        self.tree.setColumnWidth(2, 50)
        self.dialog.exec()

    def _close(self):
        self.dialog_is_open = False
        self.dialog.deleteLater()

    def _change_item(self):
        self.btn_stop.setEnabled(True if self.tree.currentItem() else False)

    def _stop_task(self):
        """
        Gracefully interrupt the specified PID validating that it actually
        belongs to a Polychromatic process.
        """
        pid = int(self.tree.currentItem().text(2))
        component = self.tree.currentItem()._component

        self.dbg.stdout("Stopping process PID " + str(pid), self.dbg.action, 1)
        process = procpid.ProcessManager(component)
        process.stop()
        self._refresh_list()

    def _reload_all(self):
        """
        Stops all processes and initiates the autostart procedures.
        """
        set_cursor_busy = shared.TabData(self.appdata).set_cursor_busy
        set_cursor_normal = shared.TabData(self.appdata).set_cursor_normal


        class RebootThread(QThread):
            @staticmethod
            def run():
                self.dbg.stdout("Now reloading background processes...", self.dbg.warning, 1)
                self.btn_stop.setEnabled(False)
                self.btn_reload.setEnabled(False)
                self.btn_reload.setText(self._("Restarting..."))
                set_cursor_busy()

                self.tree.clear()
                item = QTreeWidgetItem()
                item.setText(0, self._("Please wait..."))
                item.setIcon(0, QIcon(common.get_icon("general", "exit")))
                item.setDisabled(True)
                self.tree.addTopLevelItem(item)

                # Gracefully stop using procpid module
                procmgr = procpid.ProcessManager("helper")
                components = procmgr._get_component_pid_list()
                for component in components:
                    process = procpid.ProcessManager(component)
                    self.dbg.stdout("Stopping process PID " + str(process._get_component_pid()), self.dbg.action, 1)
                    process.stop()

                # Before starting new processes, make sure old PIDs have ceased execution
                self.dbg.stdout("Checking all processes have interrupted...", self.dbg.action, 1)
                timeout = 3
                while len(procmgr._get_component_pid_list()) > 0 and timeout > 0:
                    time.sleep(1)
                    timeout = timeout - 1
                    self.dbg.stdout("Still waiting ({0}s timeout)".format(str(timeout)), self.dbg.debug, 1)

                # Reload all by using helper's --autostart
                self.dbg.stdout("Executing autostart...", self.dbg.action, 1)
                procmgr.start_component(["--autostart"])

                # Finished
                set_cursor_normal()
                if self.dialog_is_open:
                    self.btn_reload.setEnabled(True)
                    self.btn_reload.setText(self._("Restart All"))
                    time.sleep(1)
                    self._refresh_list()

                    # Auto refresh view again if tray applet is delayed
                    tray_delay = self.appdata.preferences["tray"]["autostart_delay"]
                    time.sleep(2 if tray_delay < 2 else tray_delay + 2)
                    self._refresh_list()

        self.thread = RebootThread()
        self.thread.start()

    def _refresh_list(self):
        procmgr = procpid.ProcessManager()
        components = procmgr._get_component_pid_list()
        device_list = self.middleman.get_devices()

        tree = self.dialog.findChild(QTreeWidget, "TasksTree")
        tree.clear()
        self.btn_stop.setEnabled(False)

        for component in components:
            procmgr = procpid.ProcessManager(component)
            pid = str(procmgr._get_component_pid())
            running = procmgr.is_another_instance_is_running()

            task = self._("Playing software effect").replace("[]", component)
            task_icon = common.get_icon("effects", "play")
            if component == "tray-applet":
                task = self._("Tray Applet")
                task_icon = common.get_icon("general", "tray-applet")
            elif not running:
                task = self._("No longer running")
                task_icon = common.get_icon("general", "cancel")

            # TODO: Refactor later
            running_device = None
            for device in device_list:
                if device.serial == component:
                    running_device = device

            item = QTreeWidgetItem()
            item.setText(0, task)
            item.setIcon(0, QIcon(task_icon))
            if running_device:
                item.setText(1, running_device.name)
                item.setIcon(1, QIcon(running_device.form_factor["icon"]))
            item.setText(2, pid)
            item._component = component

            if not running:
                item.setText(2, "-")
                item.setDisabled(True)

            tree.addTopLevelItem(item)



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2020-2024 Luke Horwell <code@horwell.me>
"""
This module controls the 'Preferences' window of the Controller GUI.
"""

from ..base import PolychromaticBase
from .. import common
from .. import effects
from .. import locales
from .. import middleman
from .. import procpid
from .. import preferences as pref
from . import shared

import os
import configparser

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon, QPixmap
from PyQt5.QtWidgets import QWidget, QPushButton, QTreeWidget, QLabel, \
                            QComboBox, QCheckBox, QDialog, QSpinBox, \
                            QDoubleSpinBox, QDialogButtonBox, QTabWidget, \
                            QMessageBox, QAction, QToolButton


class PreferencesWindow(shared.TabData):
    """
    A window for adjusting the options of the application, viewing background
    processes and getting more information about backends in use.
    """
    def __init__(self, appdata):
        super().__init__(appdata)
        self.openrazer = OpenRazerPreferences(appdata)
        self.dialog = None
        self.pref_data = None
        self.prompt_restart = False
        self.restart_applet = False

        self.options = [
            # [group, item, <Qt object>, Qt object name, Inverted?]
            # -- General
            ["controller", "download_device_images", QCheckBox, "DownloadDeviceImages", False],
            ["controller", "system_qt_theme", QCheckBox, "UseSystemQtTheme", False],
            ["controller", "show_menu_bar", QCheckBox, "AlwaysHideMenuBar", True],
            ["controller", "landing_tab", QComboBox, "LandingTabCombo", False],
            ["controller", "window_behaviour", QComboBox, "WindowBehaviourCombo", False],
            ["controller", "toolbar_style", QComboBox, "ToolbarStyle", False],

            # -- Tray
            ["tray", "autostart", QCheckBox, "TrayAutoStart", False],
            ["tray", "mode", QComboBox, "TrayModeCombo", False],
            ["tray", "autostart_delay", QSpinBox, "TrayDelaySpinner", 0],

            # -- Editor
            ["editor", "live_preview", QCheckBox, "LivePreview", False],
            ["editor", "hide_key_labels", QCheckBox, "HideKeyLabels", False],
            ["editor", "system_cursors", QCheckBox, "UseSystemCursors", False],
            ["editor", "suppress_confirm_dialog", QCheckBox, "SuppressConfirmDialog", False],
            ["editor", "show_saved_colour_shades", QCheckBox, "ShowSavedColourShades", False],
        ]

    def open_window(self, open_tab=None):
        """
        Opens the Preferences window to change Polychromatic's options.

        Parameters:
            open_tab    (int)   Optionally jump to this specific tab index.
        """
        self.pref_data = pref.load_file(self.paths.preferences)
        self.prompt_restart = False
        self.restart_applet = False

        self.dialog = shared.get_ui_widget(self.appdata, "preferences", QDialog)
        self.dialog.findChild(QDialogButtonBox, "DialogButtons").accepted.connect(self._save_changes)

        # Set icons for tabs
        tabs = self.dialog.findChild(QTabWidget, "PreferencesTabs")
        tabs.setTabIcon(0, self.widgets.get_icon_qt("general", "controller"))
        tabs.setTabIcon(1, self.widgets.get_icon_qt("general", "tray-applet"))
        tabs.setTabIcon(2, self.widgets.get_icon_qt("effects", "paint"))
        tabs.setTabIcon(3, self.widgets.get_icon_qt("general", "matrix"))
        tabs.setTabIcon(4, self.widgets.get_icon_qt("emblems", "software"))

        # Set icons for controls
        if not self.appdata.system_qt_theme:
            self.dialog.findChild(QPushButton, "SavedColoursButton").setIcon(self.widgets.get_icon_qt("general", "edit"))
            self.dialog.findChild(QPushButton, "SavedColoursReset").setIcon(self.widgets.get_icon_qt("general", "reset"))

        # Options
        for option in self.options:
            self._load_option(option[0], option[1], option[2], option[3], option[4])

        self.dialog.findChild(QPushButton, "SavedColoursButton").clicked.connect(self.modify_colours)
        self.dialog.findChild(QPushButton, "SavedColoursReset").clicked.connect(self.reset_colours)

        # Create Icon Picker
        def _set_new_tray_icon(new_icon):
            self.dbg.stdout("New tray icon saved in memory: " + new_icon, self.dbg.debug, 1)
            self.pref_data["tray"]["icon"] = new_icon
            self.restart_applet = True

        tray_icon_picker = self.widgets.create_icon_picker_control(_set_new_tray_icon, self.pref_data["tray"]["icon"], self._("Choose Tray Applet Icon"), shared.IconPicker.PURPOSE_TRAY_ONLY)
        tray_icon_widget = self.dialog.findChild(QLabel, "TrayIconPickerPlaceholder")
        tray_icon_widget.parentWidget().layout().replaceWidget(tray_icon_widget, tray_icon_picker)

        # Drop custom icons when using native themes
        if self.appdata.system_qt_theme:
            combo = self.dialog.findChild(QComboBox, "LandingTabCombo")
            for i in range(0, combo.count()):
                combo.setItemIcon(i, QIcon())

        # Prompt for a restart after changing these options
        def _cb_set_restart_flag():
            self.prompt_restart = True

        self.dialog.findChild(QCheckBox, "UseSystemQtTheme").stateChanged.connect(_cb_set_restart_flag)

        # Restart the tray applet after changing these options
        def _cb_set_applet_flag(i):
            self.restart_applet = True

        self.dialog.findChild(QComboBox, "TrayModeCombo").currentIndexChanged.connect(_cb_set_applet_flag)

        # FIXME: Hide incomplete features
        self.dialog.findChild(QComboBox, "LandingTabCombo").removeItem(3)
        self.dialog.findChild(QComboBox, "LandingTabCombo").removeItem(2)

        # Disable tray applet tab if not installed
        if not procpid.ProcessManager().is_component_installed("tray-applet"):
            tabs.setTabEnabled(1, False)

        tray_hint = self.dialog.findChild(QLabel, "TrayAutoStartTip")
        tray_hint.hide()

        # Show/hide hints depending on current environment
        try:
            current_desktop = os.environ["XDG_CURRENT_DESKTOP"]

            # Hide native Qt theme hint on Qt desktops.
            if current_desktop in ["KDE", "LXQt"]:
                self.dialog.findChild(QLabel, "UseSystemQtThemeTip").hide()

            # Some environments deprecated the tray.
            for desktop in ["GNOME", "Pantheon"]:
                if current_desktop.lower().find(desktop.lower()) >= 0:
                    tray_hint.show()
        except KeyError:
            pass

        # Show time!
        self.dialog.findChild(QTabWidget, "PreferencesTabs").setCurrentIndex(open_tab if open_tab else 0)
        self.dialog.open()

    def _load_option(self, group, item, qcontrol, qid, inverted):
        """
        Applies the setting in memory with the control.

        Params:
            group           Preference group, e.g. "controller"
            item            Preference item, e.g. "landing_tab"
            qcontrol        Qt control object, e.g. QCheckBox
            qid             Name of Qt control, e.g. "UseSystemQtTheme"
            invert          (Optional, boolean) UI inverted to stored config
            must_restart    Changing this option requires application restart.
        """
        data = self.pref_data[group][item]
        widget = self.dialog.findChild(qcontrol, qid)

        if qcontrol == QCheckBox:
            widget.setChecked(not data if inverted else data)
        elif qcontrol == QComboBox:
            widget.setCurrentIndex(data)
        elif qcontrol == QSpinBox:
            widget.setValue(data)

    def _set_option(self, group, item, qcontrol, qid, inverted):
        """
        Updates the preferences in memory to reflect the UI control state.

        Params:
            Same as _get_option().
        """
        widget = self.dialog.findChild(qcontrol, qid)

        if qcontrol == QCheckBox:
            data = widget.isChecked()
            if inverted:
                data = not data
        elif qcontrol == QComboBox:
            data = widget.currentIndex()
        elif qcontrol == QSpinBox:
            data = widget.value()

        self.pref_data[group][item] = data
        self.appdata.preferences[group][item] = data

    def _save_changes(self):
        """
        Writes the preferences to file according to the GUI.
        """
        self.dbg.stdout("Saving preferences...", self.dbg.action, 1)
        for option in self.options:
            self._set_option(option[0], option[1], option[2], option[3], option[4])

        result = pref.save_file(self.paths.preferences, self.pref_data)
        if result:
            self.dbg.stdout("Save complete.", self.dbg.success, 1)
        else:
            self.dbg.stdout("Save failed! Check permissions?", self.dbg.error)
            self.widgets.open_dialog(self.widgets.dialog_error,
                                     self._("Save Error"),
                                     self._("Save failed. Please check the permissions and try again."))
            return False

        # Update memory for the rest of the application
        PolychromaticBase.preferences = self.pref_data

        # Instant reload
        if self.pref_data["controller"]["show_menu_bar"] == True:
            self.menubar.reinstate_menu_bar()
        else:
            self.menubar.hide_menu_bar()

        # Force refresh of current tab
        self.appdata.main_window.findChild(QAction, "actionRefreshTab").trigger()

        # Some options require a restart
        if self.prompt_restart:
            self.dbg.stdout("Program settings changed. Prompting to restart application.", self.dbg.action, 1)
            procmgr = procpid.ProcessManager()

            def _cb_restart_now():
                procmgr.restart_self(self.exec_path, self.exec_args)

            self.widgets.open_dialog(self.widgets.dialog_generic,
                                     self._("Restart Required"),
                                     self._("To apply these changes, the application must be restarted. Any unsaved changes will be lost."),
                                     buttons=[QMessageBox.Ok, QMessageBox.Ignore],
                                     default_button=QMessageBox.Ok,
                                     actions={QMessageBox.Ok: _cb_restart_now})

        # Reload tray applet
        if self.restart_applet:
            self.dbg.stdout("Tray applet settings changed. Will restart component.", self.dbg.success, 1)
            process = procpid.ProcessManager("tray-applet")
            process.reload()

    def modify_colours(self):
        """
        Opens the colour picker for editing saved colours for later use.
        """
        def _cb_dummy(a, b):
            # List is saved upon closing, nothing to do.
            pass

        virtual_picker = self.widgets.create_colour_control("#FFFFFF", _cb_dummy, None, self._("Saved Colours"))
        virtual_picker.findChild(QPushButton).click()

    def reset_colours(self):
        """
        Reset colours to the defaults.
        """
        def _cb_reset_colours():
            os.remove(self.paths.colours)
            pref.init(self._)

        self.widgets.open_dialog(self.widgets.dialog_generic,
                                 self._("Reset to Default Colours"),
                                 self._("All colours in the list will be reset. Continue?"),
                                 buttons=[QMessageBox.Ok, QMessageBox.Cancel],
                                 default_button=QMessageBox.Ok,
                                 actions={QMessageBox.Ok: _cb_reset_colours})


class OpenRazerPreferences(shared.TabData):
    """
    A special window for changing the OpenRazer configuration file when OpenRazer
    is in use.
    """
    def __init__(self, appdata):
        super().__init__(appdata)

        self.conf_path = None
        try:
            self.conf_path = "{0}/openrazer/razer.conf".format(os.environ["XDG_CONFIG_HOME"])
        except KeyError:
            self.conf_path = "{0}/.config/openrazer/razer.conf".format(os.environ.get("HOME", "~"))

        # Client settings
        self.client = [
            # "Filename", <data type>
            ["ripple_refresh_rate", float]
        ]

    def _get_openrazer_version(self):
        """
        Returns the current version of OpenRazer in decimal format: <major.minor>, e.g. 3.1
        """
        for backend in self.middleman.backends:
            if backend.backend_id == "openrazer":
                version = backend.version.split(".")[:2]
                return float("{0}.{1}".format(version[0], version[1]))
        return 0

    def open_log(self):
        self.menubar.openrazer.open_log()

    def restart_daemon(self):
        self.menubar.openrazer.restart_daemon()

    def open_window(self):
        """
        Opens a window for adjusting OpenRazer options.

        If configparser is not present for some reason, open the text editor.
        """
        version = self._get_openrazer_version()

        if version < 3.2:
            os.system("xdg-open file://{0}".format(self.conf_path))
            return

        self.keys = [
            # "Object Name",                "Group",    "Item",                         <data type>
            ["verbose_logging",             "General",  "verbose_logging",              bool],
            ["devices_off_on_screensaver",  "Startup",  "devices_off_on_screensaver",   bool],
            ["restore_persistence",         "Startup",  "restore_persistence",          bool],
            ["persistence_dual_boot_quirk", "Startup",  "persistence_dual_boot_quirk",  bool],
            ["battery_notifier",            "Startup",  "battery_notifier",             bool],
            ["battery_notifier_freq",       "Startup",  "battery_notifier_freq",        int], # seconds
            ["battery_notifier_percent",    "Startup",  "battery_notifier_percent",     int],
        ]

        self.dialog = shared.get_ui_widget(self.appdata, "openrazer-config", QDialog)
        self.dialog.findChild(QDialogButtonBox, "DialogButtons").accepted.connect(self._save_and_restart)

        # razer.conf
        for key in self.keys:
            object_name = key[0]
            group = key[1]
            key_name = key[2]
            data_type = key[3]

            if data_type == bool:
                chkbox = self.dialog.findChild(QCheckBox, object_name)
                chkbox.setChecked(self._read_config(group, key_name, bool))

            elif data_type == int:
                spinner = self.dialog.findChild(QSpinBox, object_name)
                spinner.setValue(self._read_config(group, key_name, int))

                if key_name == "battery_notifier_freq":
                    spinner.setValue(int(int(self._read_config(group, key_name, int) or 0) / 60))

        # Client
        for meta in self.client:
            filename = meta[0]
            data_type = meta[1]
            path = os.path.join(self.paths.config, "backends", "openrazer", filename)

            if not os.path.exists(path):
                continue

            try:
                with open(path, "r") as f:
                    data = str(f.readline()).strip()
                    data = data_type(data)
            except ValueError:
                self.dbg.stdout("Ignoring unexpected data from override file: " + filename, self.dbg.warning, 1)
                continue

            if data_type == int:
                chkbox = self.dialog.findChild(QCheckBox, filename)
                chkbox.setChecked(True if data == 1 else False)

            elif data_type == float:
                spinner = self.dialog.findChild(QDoubleSpinBox, filename)
                spinner.setValue(float(data))

        self.dialog.findChild(QCheckBox, "battery_notifier").clicked.connect(self._update_ui_state)
        self.dialog.findChild(QCheckBox, "restore_persistence").clicked.connect(self._update_ui_state)
        self._update_ui_state()

        # Hide options not available for older versions
        if version < 3.9:
            self.dialog.findChild(QWidget, "restore_persistence_related").hide()

        self.dialog.open()

    def _update_ui_state(self):
        """
        Enable/disable controls as appropriate to the currently selected choices.
        """
        battery_notifications_enabled = self.dialog.findChild(QCheckBox, "battery_notifier").isChecked()
        for object_name in ["battery_notifier_percent", "battery_notifier_freq"]:
            self.dialog.findChild(QSpinBox, object_name).setEnabled(battery_notifications_enabled)
            self.dialog.findChild(QLabel, f"{object_name}_label").setEnabled(battery_notifications_enabled)

        restore_persistence_enabled = self.dialog.findChild(QCheckBox, "restore_persistence").isChecked()
        self.dialog.findChild(QCheckBox, "persistence_dual_boot_quirk").setEnabled(restore_persistence_enabled)

    def _save_and_restart(self):
        """
        Updates the razer.conf file according to the GUI options.
        """
        # razer.conf
        for key in self.keys:
            object_name = key[0]
            group = key[1]
            key_name = key[2]
            data_type = key[3]

            if data_type == bool:
                value = self.dialog.findChild(QCheckBox, object_name).isChecked()
            elif data_type == int:
                value = self.dialog.findChild(QSpinBox, object_name).value()

                if key_name == "battery_notifier_freq":
                    value = int(value * 60)
            else:
                continue

            self._write_config(group, key_name, value)

        # Client
        for meta in self.client:
            filename = meta[0]
            data_type = meta[1]
            path = os.path.join(self.paths.config, "backends", "openrazer", filename)

            if data_type == int:
                data = 1 if self.dialog.findChild(QCheckBox, filename).isChecked() else 0
                with open(path, "w") as f:
                    f.write(str(data))

            elif data_type == float:
                data = self.dialog.findChild(QDoubleSpinBox, filename).value()
                with open(path, "w") as f:
                    f.write(str(data))

        # Restart the daemon to apply changes
        self.menubar.openrazer.restart_daemon()

    def _read_config(self, group, key_name, data_type):
        """
        Reads OpenRazer's razer.conf file, similar format to an INI.
        If value is not found, the default is False or 0.
        """
        config = configparser.ConfigParser()
        config.read(self.conf_path)

        try:
            value = config[group][key_name]
            if value == "True":
                return True
            elif value == "False":
                return False
            else:
                return data_type(value)
        except KeyError:
            # Return default data
            if data_type == int:
                return 0
            elif data_type == bool:
                return False

    def _write_config(self, group, key_name, value):
        """
        Overwrites a new key to OpenRazer's razer.conf file.
        """
        import configparser
        config = configparser.ConfigParser()
        config.read(self.conf_path)

        if group not in config:
            config[group] = {}
        config[group][key_name] = str(value)

        if not os.path.exists(os.path.dirname(self.conf_path)):
            os.makedirs(os.path.dirname(self.conf_path))

        with open(self.conf_path, "w") as f:
            config.write(f)
